@@ -116,6 +116,7 @@ unlinkFrame(FrameObj fr)
 
     ws_uncreate_frame(fr);
     deleteChain(fr->display->frames, fr);
+    deleteChain(MappedFrames, fr);
     if ( notNil(fr->application) )
       send(fr->application, NAME_delete, fr, EAV);
 
@@ -186,7 +187,8 @@ Any
 getConfirmFrame(FrameObj fr, Point pos, BoolObj grab, BoolObj normalise)
 { Any rval;
 
-  TRY( openFrame(fr, pos, grab, normalise) );
+  TRY( openFrame(fr, pos, grab, normalise) &&
+       exposeFrame(fr) );
   busyCursorDisplay(fr->display, NIL, DEFAULT);
 
   assign(fr, return_value, ConstantNotReturned);
@@ -804,7 +806,7 @@ getGeometryFrame(FrameObj fr)
     { Int n = getIndexChain(fr->display->monitors, mon);
 
       if ( n )
-	sprintf(buf+strlen(buf), "@" INTPTR_FORMAT, valInt(n)-1);
+	sprintf(buf+strlen(buf), "@%" PRIdPTR, valInt(n)-1);
     }
 
     answer(CtoName(buf));
@@ -2030,6 +2032,9 @@ makeClassFrame(Class class)
 				     NAME_NotReturned,
 				     CtoString("Used for `frame <-confirm'"),
 				     EAV);
+
+  MappedFrames = globalObject(NAME_mappedFrames, ClassChain, EAV);
+
   succeed;
 }
 
