@@ -3,8 +3,8 @@
     Author:        Jan Wielemaker
     E-mail:        J.Wielemaker@vu.nl
     WWW:           http://www.swi-prolog.org
-    Copyright (c)  1985-2022, University of Amsterdam
-                              VU University Amsterdam
+    Copyright (c)  1985-2024, University of Amsterdam
+			      VU University Amsterdam
 			      CWI, Amsterdam
 			      SWI-Prolog Solutions b.v.
     All rights reserved.
@@ -59,20 +59,19 @@
 #include "pl-trace.h"
 #include "pl-read.h"
 #include "pl-wam.h"
+#include "pl-index.h"
 #include "pl-gc.h"
 #include "pl-proc.h"
 #include "pl-pro.h"
 #include "pl-gvar.h"
 #include "pl-coverage.h"
+#include "pl-bag.h"
 #include <sys/stat.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #include <errno.h>
 
-#undef max
-#define max(a,b) ((a) > (b) ? (a) : (b))
-
 #undef K
 #undef MB
 #define K * 1024
@@ -119,6 +118,7 @@ setupProlog(void)
   initForeign();
   DEBUG(1, Sdprintf("Prolog Signal Handling ...\n"));
   initSignals();
+  initClauseIndexing();
   DEBUG(1, Sdprintf("Stacks ...\n"));
   if ( !initPrologStacks(GD->options.stackLimit) )
     outOfCore();
@@ -487,9 +487,9 @@ dispatch_signal(int sig, int sync)
 
 	  if ( PL_get_thread_alias(tid, &alias) )
 	    name = PL_atom_wchars(alias, NULL);
-	  Sdprintf("Got signal %d in thread %d (%Ws) %s\n",
-		   sig, tid, name,
-		   sync ? " (sync)" : " (async)");
+	  SdprintfX("Got signal %d in thread %d (%Ws) %s\n",
+		    sig, tid, name,
+		    sync ? " (sync)" : " (async)");
 	});
 #else
   DEBUG(MSG_SIGNAL,
@@ -669,7 +669,7 @@ prepareSignal(int sig, int plsig_flags)
     if ( !IS_VSIG(sig) )
     { handler_t old_handler = set_sighandler(sig, desired_state == PLSIG_IGNORED ? SIG_IGN : pl_signal_handler);
       if ( current_state == 0 )
-        sh->saved_handler = old_handler;
+	sh->saved_handler = old_handler;
     }
   } else
   { sh->flags = (sh->flags & ~PLSIG_USERFLAGS) | plsig_flags;
@@ -894,7 +894,7 @@ initSignals(DECL_LD)
 	sn->flags = 0;
 #endif
       if ( sn->flags )
-        prepareSignal(sn->sig, sn->flags);
+	prepareSignal(sn->sig, sn->flags);
     }
 
 #ifdef SIGHUP
@@ -914,7 +914,7 @@ initSignals(DECL_LD)
   /* these signals are not related to Unix signals and can thus */
   /* be enabled always */
 
-  PL_signal(SIG_GC|PL_SIGSYNC,	          gc_handler);
+  PL_signal(SIG_GC|PL_SIGSYNC,		  gc_handler);
   PL_signal(SIG_TUNE_GC|PL_SIGSYNC,	  gc_tune_handler);
   PL_signal(SIG_CLAUSE_GC|PL_SIGSYNC,     cgc_handler);
   PL_signal(SIG_PLABORT|PL_SIGSYNC,       abort_handler);
@@ -1373,8 +1373,8 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
 	return FALSE;
     } else
     { if ( !PL_unify_term(old, PL_FUNCTOR, FUNCTOR_colon2,
-			         PL_ATOM, def->module->name,
-			         PL_ATOM, def->functor->name) )
+				 PL_ATOM, def->module->name,
+				 PL_ATOM, def->functor->name) )
 	return FALSE;
     }
   } else if ( sh->handler )
@@ -1426,7 +1426,7 @@ PRED_IMPL("$on_signal", 4, on_signal, 0)
 
     if ( PL_get_pointer(a, &f) )
     { sh = prepareSignal(sign, 0);
-      sh->handler = (handler_t)f;
+      sh->handler = (handler_t)(intptr_t)f;
       sh->predicate = NULL;
 
       succeed;
@@ -1560,7 +1560,7 @@ init_stack(Stack s, char *name, size_t size, size_t spare, int gc)
   s->min_free   = 256*sizeof(word);
   s->max	= addPointer(s->base, size - spare);
   s->gced_size  = 0L;			/* size after last gc */
-  s->gc	        = gc;
+  s->gc		= gc;
   gcPolicy(s, GC_FAST_POLICY);
 }
 
@@ -1678,7 +1678,9 @@ gcPolicy(Stack s, int policy)
 
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-trimStacks() reclaims all unused space on the stack.
+trimStacks() reclaims all unused space on the stack. In some places this
+may be called with arguments pushed to the   next frame. In this case we
+must extend lTop to ARGP.
 - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
 
 void
@@ -1686,8 +1688,14 @@ trimStacks(DECL_LD int resize)
 { LD->trim_stack_requested = FALSE;
 
   if ( resize )
-  { growStacks(GROW_TRIM, GROW_TRIM, GROW_TRIM);
-  } else
+    growStacks(GROW_TRIM, GROW_TRIM, GROW_TRIM);
+
+  LocalFrame lSave = lTop;
+  if ( LD->query && LD->query->registers.fr && /* we have saved registers */
+       LD->query->registers.argp > (Word)lTop )
+    lTop = (LocalFrame)LD->query->registers.argp;
+
+  if ( !resize )
   { trim_stack((Stack) &LD->stacks.local);
     trim_stack((Stack) &LD->stacks.global);
     trim_stack((Stack) &LD->stacks.trail);
@@ -1707,6 +1715,8 @@ trimStacks(DECL_LD int resize)
 	{ scan_global(FALSE);
 	  checkStacks(NULL);
 	});
+
+  lTop = lSave;
 }
 
 
@@ -1741,12 +1751,7 @@ freePrologLocalData(PL_local_data_t *ld)
     destroyHTable(ld->gvar.nb_vars);
 #endif
 
-  if ( ld->bags.default_bag )
-  { PL_free(ld->bags.default_bag);
-#if defined(O_ATOMGC) && defined(O_PLMT)
-    simpleMutexDelete(&ld->bags.mutex);
-#endif
-  }
+  cleanup_bags(ld);
 
 #ifdef O_CYCLIC
   clearSegStack(&ld->cycle.lstack);
@@ -1831,14 +1836,14 @@ set_stack_limit(size_t limit)
 
   if ( limit < LD->stacks.limit &&
        limit < sizeStack(local) +
-               sizeStack(global) +
-               sizeStack(trail) )
+	       sizeStack(global) +
+	       sizeStack(trail) )
   { garbageCollect(GC_USER);
     trimStacks(TRUE);
 
     if ( limit < sizeStack(local) +
-	         sizeStack(global) +
-	         sizeStack(trail) )
+		 sizeStack(global) +
+		 sizeStack(trail) )
     { term_t ex;
 
 
